/***************************************************************************
 *   Copyright (C) 2004 by Joris Guisson                                   *
 *   joris.guisson@student.luc.ac.be                                       *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <qstring.h>
#include <qfileinfo.h>
#include "util/error.h"
#include "gstplaythread.h"
#include "globals.h"

#ifdef HAVE_GSTREAMER

void eos(GstElement *element, gpointer data) 
{
	GstPlayThread* pl = (GstPlayThread*)data;
	pl->m_playing = false;
}

void error_callback(GObject* object,GstObject* orig,gchar* error)
{
	OutputLog() << "Error " << GST_OBJECT_NAME (orig) << " : " << error << endl;
}

GstPlayThread::GstPlayThread()
		: QThread()
{
	m_pipeline = m_volume = m_audiosink = 0;
	m_playing = false;
	m_paused = false;
	m_volume_val = 1.0f;
}


GstPlayThread::~GstPlayThread()
{
	/* stop the pipeline */
	if (m_pipeline)
	{
		gst_element_set_state (m_pipeline, GST_STATE_NULL);
		gst_object_unref (GST_OBJECT (m_pipeline));
	}
}

void GstPlayThread::buildPipeline()
{
	GstElement* pipeline = gst_pipeline_new ("pipeline");
	GstElement* thread = gst_thread_new ("thread");
	GstElement* queue = gst_element_factory_make ("queue", "queue");
	GstElement* decoder = gst_element_factory_make ("spider", "decoder");
	GstElement* filesrc = gst_element_factory_make ("filesrc", "disk_source");
	GstElement* volume = gst_element_factory_make("volume", "volume");
	GstElement* audiosink = gst_element_factory_make ("artssink", "play_audio");
	
	if (!pipeline || !decoder || !filesrc || !volume || !audiosink)
		throw Error("Cannot create GStreamer backend !");
	
	g_signal_connect(pipeline,"error",G_CALLBACK(error_callback),NULL);
	g_signal_connect (G_OBJECT (filesrc), "eos",G_CALLBACK (eos), this);
	
	gst_bin_add_many (GST_BIN (thread), volume, audiosink, NULL);	
	gst_bin_add_many (GST_BIN (pipeline), filesrc,decoder,queue, thread, NULL);	
	gst_element_link_many(filesrc,decoder,queue,volume, audiosink,NULL);
	
	
	m_pipeline = pipeline;
	m_volume = volume;
	m_audiosink = audiosink;
	m_filesrc = filesrc;
	m_decoder = decoder;
	m_queue = queue;
	setVolume(m_volume_val);
}

void GstPlayThread::reconfigurePipeline()
{
	gst_object_unref (GST_OBJECT (m_pipeline));
	buildPipeline();
//	qDebug("Reconfiguring pipeline");
}
		
float GstPlayThread::setVolume(float vol)
{
	if (!m_volume)
		return vol;
	
	if (vol < 0)
		vol = 0;
	if (vol > 1)
		vol = 1;
	
	g_object_set (G_OBJECT (m_volume), "volume", (gdouble)vol, NULL);
	m_volume_val = vol;
	return vol;
}

void GstPlayThread::run()
{
	while (m_playing)
		gst_bin_iterate(GST_BIN(m_pipeline));
}

void GstPlayThread::play(const char* name)
{
	if (!m_pipeline)
	{
		buildPipeline();
	}
	
	if (m_file_type == QString::null || 
		   QFileInfo(m_file_type).extension() == QFileInfo(name).extension())
	{
		g_object_set (G_OBJECT (m_filesrc), "location", name, NULL);
	}
	else
	{
		reconfigurePipeline();
		g_object_set (G_OBJECT (m_filesrc), "location", name, NULL);
	}
	
	m_file_type = name;
	m_playing = true;
	gst_element_set_state(m_pipeline, GST_STATE_PLAYING);
}

void GstPlayThread::pause()
{
	if (m_playing && !m_paused)
	{
		gst_element_set_state (m_pipeline,GST_STATE_PAUSED);
		m_paused = true;
	}
}

void GstPlayThread::resume()
{
	if (m_playing && m_paused)
	{
		gst_element_set_state(m_pipeline, GST_STATE_PLAYING);
		m_paused = false;
	}
}

void GstPlayThread::stop()
{
	if (m_playing)
	{
		gst_element_set_state(m_pipeline, GST_STATE_NULL);
		m_playing = false;
		m_paused = false;
	}
}

void GstPlayThread::seek(int secs)
{
	if (!m_pipeline)
		return;
	
	int type = (GST_FORMAT_TIME | GST_SEEK_METHOD_SET | GST_SEEK_FLAG_FLUSH);
	gst_element_seek(m_audiosink, GstSeekType(type), secs * GST_SECOND);
	
	//int st = GST_SEEK_METHOD_SET;
	//gst_element_seek(m_pipeline, (GstSeekType)st, secs * GST_SECOND);
}

int GstPlayThread::currentTime() const
{
	if (!m_pipeline)
		return 0;
	
	gint64 ns = 0;
	GstFormat format = GST_FORMAT_TIME;
	gst_element_query(m_audiosink, GST_QUERY_POSITION, &format, &ns);
	return ns / GST_SECOND;
}

#endif
