/***************************************************************************
 *   Copyright (C) 2004 by Joris Guisson                                   *
 *   joris.guisson@pandora.be                                       *
 *                                                                         *
 ***************************************************************************/
#include <qtimer.h>
#include <qthread.h>
#include <kconfig.h>
#include <ktoolbar.h>
#include <kmessagebox.h>
#include "musicplayer.h"
#include "musicdb/song.h"
#include "songprovider.h"
#include "playhistory.h"
#include "playqueue.h"
#include "playerfactoryrepository.h"
#include "arts/artsplayerfactory.h"
#include "gstreamer/gstplayerfactory.h"
#include "musicdb/musicdatabase.h"
#include "util/error.h"
#include "globals.h"





MusicPlayer::MusicPlayer(MusicDatabase* mdb,QObject *parent, const char *name)
		: QObject(parent, name),
		m_player(0),m_db(0),m_repeat(false),m_random(false),m_volume(1.0f)
{
	m_pfrep = new PlayerFactoryRepository();
	m_pfrep->addFactory(new ArtsPlayerFactory());
//#ifdef HAVE_GSTREAMER
//	m_pfrep->addFactory(new GstPlayerFactory());
//#endif
	m_backend_name = "arts";
	m_player = m_pfrep->getFactory(m_backend_name)->createPlayer();
	m_timer = new QTimer(this);
	
	connect(m_timer,SIGNAL(timeout()),this,SLOT(onTimeout()));
	setVolume(m_volume);
	setObjId("player");
	
	m_history = new PlayHistory();
	m_play_queue = new PlayQueue(mdb);
}


MusicPlayer::~MusicPlayer()
{
	delete m_player;
	delete m_history;
	delete m_play_queue;
	delete m_pfrep;
}

void MusicPlayer::changeBackend(const QString & name)
{
	OutputLog() << "Changing backend to " << name << endl;
	if (m_backend_name == name)
		return;
	
	PlayerFactory* pf = m_pfrep->getFactory(name);
	if (!pf)
		return;
	
	PlayerBase* np = 0;
	try
	{
		np = pf->createPlayer();
	}
	catch (Error & e)
	{
		KMessageBox::error(0,"Error : " + e.errorMessage(),"Error");
		return;
	}
	
	if (m_player->isPlaying())
	{
		int ct = m_player->currentTime();
		Song* s = m_history->top();
		QString file;
		if (s)
			file = s->getFilename();
		
		m_player->stop();
		delete m_player;
		
		m_player = np;
		m_volume = m_player->setVolume(m_volume);
		if (file != QString::null && m_player->play(file))
		{
			m_player->seek(ct);
		}
		m_backend_name = name;
	}
	else
	{
		delete m_player;
		m_player = np;
		m_volume = m_player->setVolume(m_volume);
		m_backend_name = name;
	}
}

void MusicPlayer::setProvider(SongProvider* db)
{
	if (db)
		m_db = db;
}


void MusicPlayer::onTimeout()
{
	if (m_player->isIdle())
		playNextSong(true);
	
	elapsed(m_player->currentTime());
}

void MusicPlayer::enqueueSong(Song* s)
{
	m_play_queue->enqueue(s);
}

void MusicPlayer::addToHistory(Song* fn)
{
	m_history->push(fn);
	setPrevEnabled(true);
}

Song* MusicPlayer::popHistory()
{
	Song* s = m_history->pop();
	if (m_history->empty())
		setPrevEnabled(false);
	return s;
}

void MusicPlayer::nothingPlaying()
{
	playing("Nothing playing");
	totalSongTimeChanged(0);
	if (m_timer->isActive())
		m_timer->stop();
}

void MusicPlayer::playNextSong(bool at_end_of_song)
{
	Song* s = 0;
	
	
	if (!m_play_queue->empty())
	{
		s = m_play_queue->dequeue();
	}
	else if (m_random || m_player->isPlaying() || at_end_of_song)
	{
		s = m_db->nextSong(m_random,m_repeat);
	}
	else
	{
		s = m_db->firstSong();
	}
	
	stop();

	if (!s)
	{
		nothingPlaying();
	}
	else
	{
		QString file = s->getFilename();
		
		if (m_player->play(file))
		{
			addToHistory(s);
			playing(file);
			totalSongTimeChanged(s->getLength());
			elapsed(0);
			if (!m_timer->isActive())
				m_timer->start(1000);
			setPauseEnabled(true);
			setPlayEnabled(false);
			setStopEnabled(true);
		}
	}
}

void MusicPlayer::play()
{	
	if (m_player->isPaused())
	{
		m_player->resume();
		if (!m_timer->isActive())
			m_timer->start(1000);
		setPauseEnabled(true);
		setPlayEnabled(false);
		setStopEnabled(true);
	}
	else
	{
		playNextSong();
	}
}

void MusicPlayer::pause()
{
	if (m_player->pause())
	{
		if (m_timer->isActive())
			m_timer->stop();
		setPauseEnabled(false);
		setPlayEnabled(true);
		setStopEnabled(true);
	}
}

void MusicPlayer::stop()
{
	if (m_player->stop())
	{
		nothingPlaying();
		setPauseEnabled(false);
		setPlayEnabled(true);
		setStopEnabled(false);
	}
}

void MusicPlayer::next()
{
	playNextSong();
}

void MusicPlayer::prev()
{
	stop();
	
	Song* last = popHistory();
	if (!last)
		return;
	
	if (m_player->play(last->getFilename()))
	{
		addToHistory(last);
		totalSongTimeChanged(last->getLength());
		elapsed(0);
		playing(last->getFilename());
		if (!m_timer->isActive())
			m_timer->start(1000);
		
		setPauseEnabled(true);
		setPlayEnabled(false);
		setStopEnabled(true);
	}
}

void MusicPlayer::setRandom(bool on)
{
	m_random = on;
}

void MusicPlayer::setRepeat(bool on)
{
	m_repeat = on;
}

void MusicPlayer::seek(int secs)
{
	m_player->seek(secs);
}

void MusicPlayer::playSong(Song* s)
{
	stop();
	if (!s)
		return;
	
	QString file = s->getFilename();
	
	if (m_player->play(file))
	{
		addToHistory(s);
		playing(file);
		totalSongTimeChanged(s->getLength());
		elapsed(0);
		if (!m_timer->isActive())
			m_timer->start(1000);
		setPauseEnabled(true);
		setPlayEnabled(false);
		setStopEnabled(true);
	}
}

void MusicPlayer::setVolume(float vol)
{
	m_volume = m_player->setVolume(vol);
	volumeChanged(m_volume);
}

void MusicPlayer::saveSettings(KConfig* cfg)
{
	cfg->setGroup("player");
	cfg->writeEntry("random",m_random);
	cfg->writeEntry("repeat",m_repeat);
	cfg->writeEntry("volume",(double)m_volume);
	cfg->writeEntry("backend",m_backend_name);
}

void MusicPlayer::loadSettings(KConfig* cfg)
{
	cfg->setGroup("player");
	float v = cfg->readDoubleNumEntry("volume",1.0);
	setVolume(v);
	
	m_repeat = cfg->readBoolEntry("repeat",false);
	m_random = cfg->readBoolEntry("random",false);
	changeBackend(cfg->readEntry("backend","arts"));
}

#include "musicplayer.moc"
